@@ -22,23 +22,22 @@
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 import argparse
-import bz2
 import fnmatch
-import gzip
-import io
 import os
 import re
 import shlex
 import socket
 import time
+import subprocess
 
 from copy import copy
 from calendar import timegm
-from datetime import datetime
+from datetime import datetime, timedelta
 from math import log10, exp, sqrt
 
 from flent.loggers import get_logger
 
+MULTIHOST_SEP = '//'
 ENCODING = "UTF-8"
 try:
     import locale
@@ -85,21 +84,61 @@ def format_bytes(nbytes):
         return (nbytes, 'bytes')
 
 
-def parse_date(timestring):
+def parse_date(timestring, min_t=None, offset=None):
     try:
         # Try to parse the straight UTC time string (has a Z at the end)
-        return datetime.strptime(timestring, "%Y-%m-%dT%H:%M:%S.%fZ")
+        return datetime.strptime(timestring, "%Y-%m-%dT%H:%M:%S.%fZ"), None
     except ValueError:
         try:
             dt = datetime.strptime(timestring, "%Y-%m-%dT%H:%M:%S.%f")
         except ValueError:
             dt = datetime.strptime(timestring, "%Y-%m-%dT%H:%M:%S")
-        # The timestamp is in local time, so get the (UTC) timestamp and
-        # subtract the time zone offset at that time to get the UTC datetime
-        # object.
-        ts = time.mktime(dt.timetuple())
-        offset = datetime.fromtimestamp(ts) - datetime.utcfromtimestamp(ts)
-        return dt - offset
+
+        if offset is None:
+            # The timestamp is in local time, so we need to find the timezone
+            # offset. Guess this by computing the nearest whole-hour offset for
+            # from the minimum timestamp in the data series (passed as min_t)
+            #
+            # If not min_t is available, compute the offset from the system time
+            # zone.
+            #
+            # Return the offset so it can be re-used the next time this function
+            # is called.
+
+            logger.debug("Non-UTC timestamp %s, computing timezone offset",
+                         timestring)
+
+            if min_t is not None:
+                ts = orig_ts = timegm(dt.timetuple())
+                hours = 0
+                mult = 1 if ts < min_t else -1
+                while abs(ts - min_t) > 1800:
+                    hours -= 1 * mult
+                    ts += 3600 * mult
+                offset = timedelta(hours=hours)
+                logger.debug("Guessed timezone offset of %d hours for "
+                             "ts %f min_t %f diff %f",
+                             hours, orig_ts, min_t, orig_ts-min_t)
+            else:
+                ts = time.mktime(dt.timetuple())
+                offset = (datetime.fromtimestamp(ts) -
+                          datetime.utcfromtimestamp(ts))
+                logger.debug("Computed offset of %s from system timezone",
+                             offset)
+
+        return dt - offset, offset
+
+
+def parse_int(val):
+    try:
+        try:
+            return int(val)
+        except ValueError:
+            if val.strip().startswith("0x"):
+                return int(val, 16)
+            raise
+    except (ValueError, AttributeError):
+        raise ValueError("Invalid integer value: %s" % val)
 
 
 def clean_path(path, allow_dirs=False):
@@ -142,12 +181,24 @@ def is_executable(filename):
     return os.path.isfile(filename) and os.access(filename, os.X_OK)
 
 
-def which(executable, fail=None):
+def which(executable, fail=None, remote_host=None):
     pathname, filename = os.path.split(executable)
     if pathname:
         if is_executable(executable):
             logger.debug("which: %s is a full path and executable", executable)
             return executable
+    elif remote_host:
+        logger.debug("running 'which' for binary '%s' on host '%s'",
+                     executable, remote_host)
+        try:
+            output = subprocess.check_output(['ssh', remote_host,
+                                              'which {}'.format(executable)],
+                                             timeout=1)
+            output = output.decode(ENCODING).strip()
+            logger.debug("Got path '%s' for '%s' on '%s'", output, executable, remote_host)
+            return output
+        except subprocess.CalledProcessError:
+            pass
     else:
         for path in [i.strip('""') for i in os.environ["PATH"].split(os.pathsep)]:
             filename = os.path.join(path, executable)
@@ -174,8 +225,13 @@ def path_components(path):
         folders.insert(0, path)
     return folders
 
+def normalise_host(hostname):
+    if hostname and MULTIHOST_SEP in hostname:
+        return hostname.split(MULTIHOST_SEP, 1)[0]
+    return hostname
 
 def lookup_host(hostname, version=None):
+    hostname = normalise_host(hostname)
     logger.debug("Looking up hostname '%s'.", hostname)
     if version == 4:
         version = socket.AF_INET
@@ -195,82 +251,6 @@ def lookup_host(hostname, version=None):
     return hostnames[0]
 
 
-# In Python 2.6, the GzipFile object does not have a 'closed' property, which
-# makes the io module blow up when trying to close it. This tidbit tries to
-# detect that and substitute a subclass that does have the property, while not
-# touching anything if the property is already present.
-if hasattr(gzip.GzipFile, "closed"):
-    _gzip_open = gzip.open
-else:
-    class GzipFile(gzip.GzipFile):
-
-        def get_closed(self):
-            return self.fileobj is None
-        # Setter needed for python3.1-compatibility
-
-        def set_closed(self, closed):
-            self._closed = closed
-        closed = property(get_closed, set_closed)
-    _gzip_open = GzipFile
-
-
-def gzip_open(filename, mode="rb"):
-    """Compatibility layer for gzip to work in Python 3.1 and 3.2."""
-    wrap_text = False
-    if "t" in mode:
-        wrap_text = True
-        mode = mode.replace("t", "")
-    binary_file = _gzip_open(filename, mode)
-
-    if wrap_text:
-        # monkey-patching required to make gzip object compatible with
-        # TextIOWrapper in Python 3.1.
-        if not hasattr(binary_file, "readable"):
-            def readable():
-                return binary_file.mode == gzip.READ
-            binary_file.readable = readable
-        if not hasattr(binary_file, "writable"):
-            def writable():
-                return binary_file.mode == gzip.WRITE
-            binary_file.writable = writable
-        if not hasattr(binary_file, "seekable"):
-            def seekable():
-                return True
-            binary_file.seekable = seekable
-
-        # This wrapping is done by the builtin gzip module in python 3.3.
-        return io.TextIOWrapper(binary_file)
-    else:
-        return binary_file
-
-
-if hasattr(bz2, 'open'):
-    bz2_open = bz2.open
-else:
-    # compatibility with io.TextIOWrapper for Python 2
-    class bz2file(bz2.BZ2File):
-
-        def readable(self):
-            return 'r' in self.mode
-
-        def writable(self):
-            return 'w' in self.mode
-
-        def seekable(self):
-            return True
-
-        def flush(self):
-            pass
-
-    def bz2_open(filename, mode='rb', compresslevel=9):
-        bz_mode = mode.replace("t", "")
-        binary_file = bz2file(filename, bz_mode, compresslevel=compresslevel)
-        if "t" in mode:
-            return io.TextIOWrapper(binary_file)
-        else:
-            return binary_file
-
-
 class DefaultConfigParser(configparser.ConfigParser):
 
     class _NoDefault(object):
@@ -333,10 +313,9 @@ class Glob(object):
             pattern = self.pattern
 
         # Exclude * from matching :, make ** match everything
-        re_pat = fnmatch.translate(pattern)
-        re_pat = re_pat.replace(".*.*", ":::PLACEHOLDER:::")
+        re_pat = fnmatch.translate(pattern.replace("**", "___PLACEHOLDER___"))
         re_pat = re_pat.replace(".*", "[^:]*")
-        re_pat = re_pat.replace(":::PLACEHOLDER:::", ".*")
+        re_pat = re_pat.replace("___PLACEHOLDER___", ".*")
         regex = re.compile(re_pat)
 
         exclude += self.exclude
@@ -422,7 +401,7 @@ def mos_score(T, loss):
         Idd = 0
     else:
         Idd = 25 * ((1 + X**6)**(1 / 6) - 3 *
-                    (1 + (X / 3)**6)**1 / 6 + 2)  # (7-27)
+                    (1 + (X / 3)**6)**(1 / 6) + 2)  # (7-27)
 
     Idle = (Ro - Rle) / 2 + sqrt((Ro - Rle)**2 / 4 + 169)  # (7-25)
 
@@ -478,6 +457,44 @@ class Update(argparse.Action):
         getattr(namespace, self.dest).update(values)
 
 
+def _copy_items(items):
+    if items is None:
+        return []
+    # The copy module is used only in the 'append' and 'append_const'
+    # actions, and it is needed only when the default value isn't a list.
+    # Delay its import for speeding up the common case.
+    if type(items) is list:
+        return items[:]
+    return copy(items)
+
+def append_host(items, new_host):
+    host, suffix, idx = (None, None, None)
+    for k, v in enumerate(items):
+        if MULTIHOST_SEP in v:
+            h, s = v.split(MULTIHOST_SEP, 1)
+            s = int(s)
+        else:
+            h, s = (v, 1)
+        if h == new_host:
+            host, suffix, idx = h, s, k
+    if host:
+        new_host = MULTIHOST_SEP.join((new_host, str(suffix+1)))
+        items[idx] = MULTIHOST_SEP.join((host, str(suffix)))
+    items.append(new_host)
+
+class AddHost(argparse._AppendAction):
+
+    def __init__(self, *args, **kwargs):
+        if 'default' not in kwargs:
+            kwargs['default'] = []
+        super(AddHost, self).__init__(*args, **kwargs)
+
+    def __call__(self, parser, namespace, new_host, option_string=None):
+        items = getattr(namespace, self.dest, None)
+        items = _copy_items(items)
+        append_host(items, new_host)
+        setattr(namespace, self.dest, items)
+
 def float_pair(value):
     try:
         if "," not in value:
@@ -489,6 +506,13 @@ def float_pair(value):
         raise argparse.ArgumentTypeError("Invalid pair value: %s" % value)
 
 
+def float_pair_noomit(value):
+    val = float_pair(value)
+    if None in val:
+        raise argparse.ArgumentTypeError("Invalid pair value: %s" % value)
+    return val
+
+
 def keyval(value):
     ret = {}
     for p in token_split(value, ";"):
@@ -503,12 +527,44 @@ def keyval(value):
                 "Invalid value '%s' (missing =)" % p)
     return ret
 
+def noop(x):
+    return x
 
-def keyval_int(value):
-    try:
-        return {int(k): v for k, v in keyval(value).items()}
-    except ValueError:
-        raise argparse.ArgumentTypeError("Keys must be integers.")
+def todict(k, v):
+    return dict(k=v)
+
+def rangedict(key, value):
+    ret = {}
+    for k in key.split(","):
+        if '-' in k:
+            s, e = (int(i) for i in k.split("-", 1))
+            if not s < e:
+                raise ValueError
+            for i in range(s, e+1):
+                ret[i] = value
+        elif k == '*':
+            ret['*'] = value
+        else:
+            ret[int(k)] = value
+    return ret
+
+def keyval_pair_transformer(pairfunc=todict, errmsg="Parse error"):
+    def typefunc(value):
+        ret = {}
+        try:
+            for k, v in keyval(value).items():
+                ret.update(pairfunc(k, v))
+            return ret
+        except ValueError:
+            raise argparse.ArgumentTypeError(errmsg)
+    return typefunc
+
+def keyval_transformer(keyfunc=noop, valfunc=noop, errmsg="Parse error"):
+    return keyval_pair_transformer(pairfunc=lambda k, v: {keyfunc(k): valfunc(v)},
+                                   errmsg=errmsg)
+
+keyval_int = keyval_pair_transformer(pairfunc=rangedict,
+                                     errmsg="Keys must be integers.")
 
 
 def comma_list(value):
